import re
from datetime import timedelta
from enum import Enum
from pathlib import Path
from typing import Iterable, Set, Dict

from compal.functions import GetFunction as GET
from lxml import etree
from prometheus_client import Metric
from prometheus_client.metrics_core import (
    InfoMetricFamily,
    CounterMetricFamily,
    GaugeMetricFamily,
    StateSetMetricFamily,
)

DEVICE_STATUS = "device_status"
DOWNSTREAM = "downstream"
UPSTREAM = "upstream"
LAN_USERS = "lan_users"
TEMPERATURE = "temperature"


class XmlMetricsExtractor:

    PROJECT_ROOT = Path(__file__).parent.parent
    SCHEMA_ROOT = PROJECT_ROOT / "resources" / "schema"

    def __init__(self, name: str, functions: Set[int]):
        self._name = name

        # create one parser per function, use an XML schema if available
        self._parsers = {}
        for fun in functions:
            path = XmlMetricsExtractor.SCHEMA_ROOT / f"{fun}.xsd"
            if path.exists():
                schema = etree.XMLSchema(file=str(path))
            else:
                schema = None
            parser = etree.XMLParser(schema=schema)
            self._parsers[fun] = parser

    @property
    def name(self):
        """
        Descriptive name for this extractor, to be used in metric labels
        :return:
        """
        return self._name

    @property
    def functions(self) -> Iterable[int]:
        """
        Connect Box getter.xml function(s) this metrics extractor is working on
        :return:
        """
        return self._parsers.keys()

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        """
        Returns metrics given raw XML responses corresponding to the functions returned in the `functions` property.
        :param raw_xmls:
        :return: metrics iterable
        :raises: lxml.etree.XMLSyntaxError in case a raw XML does not match the expected schema
        """
        raise NotImplementedError


class DownstreamStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(DownstreamStatusExtractor, self).__init__(
            DOWNSTREAM, {GET.DOWNSTREAM_TABLE, GET.SIGNAL_TABLE}
        )

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 2

        # DOWNSTREAM_TABLE gives us frequency, power level, SNR and RxMER per channel
        root = etree.fromstring(
            raw_xmls[GET.DOWNSTREAM_TABLE], parser=self._parsers[GET.DOWNSTREAM_TABLE]
        )

        CHANNEL_ID = "channel_id"
        ds_frequency = GaugeMetricFamily(
            "connectbox_downstream_frequency",
            "Downstream channel frequency",
            unit="hz",
            labels=[CHANNEL_ID],
        )
        ds_power_level = GaugeMetricFamily(
            "connectbox_downstream_power_level",
            "Downstream channel power level",
            unit="dbmv",
            labels=[CHANNEL_ID],
        )
        ds_snr = GaugeMetricFamily(
            "connectbox_downstream_snr",
            "Downstream channel signal-to-noise ratio (SNR)",
            unit="db",
            labels=[CHANNEL_ID],
        )
        ds_rxmer = GaugeMetricFamily(
            "connectbox_downstream_rxmer",
            "Downstream channel receive modulation error ratio (RxMER)",
            unit="db",
            labels=[CHANNEL_ID],
        )
        for channel in root.findall("downstream"):
            channel_id = channel.find("chid").text
            frequency = int(channel.find("freq").text)
            power_level = int(channel.find("pow").text)
            snr = int(channel.find("snr").text)
            rxmer = float(channel.find("RxMER").text)

            labels = [channel_id.zfill(2)]
            ds_frequency.add_metric(labels, frequency)
            ds_power_level.add_metric(labels, power_level)
            ds_snr.add_metric(labels, snr)
            ds_rxmer.add_metric(labels, rxmer)
        yield from [ds_frequency, ds_power_level, ds_snr, ds_rxmer]

        # SIGNAL_TABLE gives us unerrored, corrected and uncorrectable errors per channel
        root = etree.fromstring(
            raw_xmls[GET.SIGNAL_TABLE], parser=self._parsers[GET.SIGNAL_TABLE]
        )

        ds_unerrored_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_unerrored",
            "Unerrored downstream codewords",
            labels=[CHANNEL_ID],
        )
        ds_correctable_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_corrected",
            "Corrected downstream codewords",
            labels=[CHANNEL_ID],
        )
        ds_uncorrectable_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_uncorrectable",
            "Uncorrectable downstream codewords",
            labels=[CHANNEL_ID],
        )
        for channel in root.findall("signal"):
            channel_id = channel.find("dsid").text
            unerrored = int(channel.find("unerrored").text)
            correctable = int(channel.find("correctable").text)
            uncorrectable = int(channel.find("uncorrectable").text)

            labels = [channel_id.zfill(2)]
            ds_unerrored_codewords.add_metric(labels, unerrored)
            ds_correctable_codewords.add_metric(labels, correctable)
            ds_uncorrectable_codewords.add_metric(labels, uncorrectable)
        yield from [
            ds_unerrored_codewords,
            ds_correctable_codewords,
            ds_uncorrectable_codewords,
        ]


class UpstreamStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(UpstreamStatusExtractor, self).__init__(UPSTREAM, {GET.UPSTREAM_TABLE})

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1

        CHANNEL_ID = "channel_id"
        TIMEOUT_TYPE = "timeout_type"

        us_frequency = GaugeMetricFamily(
            "connectbox_upstream_frequency",
            "Upstream channel frequency",
            unit="hz",
            labels=[CHANNEL_ID],
        )
        us_power_level = GaugeMetricFamily(
            "connectbox_upstream_power_level",
            "Upstream channel power level",
            unit="dbmv",
            labels=[CHANNEL_ID],
        )
        us_symbol_rate = GaugeMetricFamily(
            "connectbox_upstream_symbol_rate",
            "Upstream channel symbol rate",
            unit="ksps",
            labels=[CHANNEL_ID],
        )
        us_timeouts = CounterMetricFamily(
            "connectbox_upstream_timeouts",
            "Upstream channel timeout occurrences",
            labels=[CHANNEL_ID, TIMEOUT_TYPE],
        )
        root = etree.fromstring(
            raw_xmls[GET.UPSTREAM_TABLE], parser=self._parsers[GET.UPSTREAM_TABLE]
        )
        for channel in root.findall("upstream"):
            channel_id = channel.find("usid").text

            frequency = int(channel.find("freq").text)
            power_level = float(channel.find("power").text)
            symbol_rate = float(channel.find("srate").text)
            t1_timeouts = int(channel.find("t1Timeouts").text)
            t2_timeouts = int(channel.find("t2Timeouts").text)
            t3_timeouts = int(channel.find("t3Timeouts").text)
            t4_timeouts = int(channel.find("t4Timeouts").text)

            labels = [channel_id.zfill(2)]
            us_frequency.add_metric(labels, frequency)
            us_power_level.add_metric(labels, power_level)
            us_symbol_rate.add_metric(labels, symbol_rate)
            us_timeouts.add_metric(labels + ["T1"], t1_timeouts)
            us_timeouts.add_metric(labels + ["T2"], t2_timeouts)
            us_timeouts.add_metric(labels + ["T3"], t3_timeouts)
            us_timeouts.add_metric(labels + ["T4"], t4_timeouts)
        yield from [us_frequency, us_power_level, us_symbol_rate, us_timeouts]


class LanUserExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(LanUserExtractor, self).__init__(LAN_USERS, {GET.LANUSERTABLE})

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1

        root = etree.fromstring(
            raw_xmls[GET.LANUSERTABLE], parser=self._parsers[GET.LANUSERTABLE]
        )

        # LAN and Wi-Fi clients have the same XML format so we can reuse the code to extract their values
        def extract_client(client, target_metric: GaugeMetricFamily):
            mac_address = client.find("MACAddr").text

            # depending on the firmware, both IPv4/IPv6 addresses or only one of both are reported
            ipv4_address_elmt = client.find("IPv4Addr")
            ipv4_address = (
                ipv4_address_elmt.text if ipv4_address_elmt is not None else ""
            )
            ipv6_address_elmt = client.find("IPv6Addr")
            ipv6_address = (
                ipv6_address_elmt.text if ipv6_address_elmt is not None else ""
            )

            hostname = client.find("hostname").text
            speed = int(client.find("speed").text)
            target_metric.add_metric(
                [mac_address, ipv4_address, ipv6_address, hostname], speed
            )

        label_names = ["mac_address", "ipv4_address", "ipv6_address", "hostname"]

        # set up ethernet user speed metric
        ethernet_user_speed = GaugeMetricFamily(
            "connectbox_ethernet_client_speed",
            "Ethernet client network speed",
            labels=label_names,
            unit="mbit",
        )
        for client in root.find("Ethernet").findall("clientinfo"):
            extract_client(client, ethernet_user_speed)
        yield ethernet_user_speed

        # set up Wi-Fi user speed metric
        wifi_user_speed = GaugeMetricFamily(
            "connectbox_wifi_client_speed",
            "Wi-Fi client network speed",
            labels=label_names,
            unit="mbit",
        )
        for client in root.find("WIFI").findall("clientinfo"):
            extract_client(client, wifi_user_speed)
        yield wifi_user_speed


class TemperatureExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(TemperatureExtractor, self).__init__(TEMPERATURE, {GET.CMSTATE})

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1

        root = etree.fromstring(
            raw_xmls[GET.CMSTATE], parser=self._parsers[GET.CMSTATE]
        )

        fahrenheit_to_celsius = lambda f: (f - 32) * 5.0 / 9
        tuner_temperature = fahrenheit_to_celsius(
            float(root.find("TunnerTemperature").text)
        )
        temperature = fahrenheit_to_celsius(float(root.find("Temperature").text))

        yield GaugeMetricFamily(
            "connectbox_tuner_temperature",
            "Tuner temperature",
            unit="celsius",
            value=tuner_temperature,
        )
        yield GaugeMetricFamily(
            "connectbox_temperature", "Temperature", unit="celsius", value=temperature,
        )


class ProvisioningStatus(Enum):
    ONLINE = "Online"
    PARTIAL_SERVICE_US = "Partial Service (US only)"
    PARTIAL_SERVICE_DS = "Partial Service (DS only)"
    PARTIAL_SERVICE_USDS = "Partial Service (US+DS)"


class DeviceStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(DeviceStatusExtractor, self).__init__(
            DEVICE_STATUS, {GET.GLOBALSETTINGS, GET.CM_SYSTEM_INFO, GET.CMSTATUS}
        )

    def extract(self, raw_xmls: Dict[int, bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 3

        # parse GlobalSettings
        root = etree.fromstring(
            raw_xmls[GET.GLOBALSETTINGS], parser=self._parsers[GET.GLOBALSETTINGS]
        )
        firmware_version = root.find("SwVersion").text
        cm_provision_mode = root.find("CmProvisionMode").text
        gw_provision_mode = root.find("GwProvisionMode").text
        operator_id = root.find("OperatorId").text

        # parse cm_system_info
        root = etree.fromstring(
            raw_xmls[GET.CM_SYSTEM_INFO], parser=self._parsers[GET.CM_SYSTEM_INFO]
        )
        docsis_mode = root.find("cm_docsis_mode").text
        hardware_version = root.find("cm_hardware_version").text
        uptime_as_str = root.find("cm_system_uptime").text

        # parse cmstatus
        root = etree.fromstring(
            raw_xmls[GET.CMSTATUS], parser=self._parsers[GET.CMSTATUS]
        )
        cable_modem_status = root.find("cm_comment").text
        provisioning_status = root.find("provisioning_st").text

        yield InfoMetricFamily(
            "connectbox_device",
            "Assorted device information",
            value={
                "hardware_version": hardware_version,
                "firmware_version": firmware_version,
                "docsis_mode": docsis_mode,
                "cm_provision_mode": cm_provision_mode,
                "gw_provision_mode": gw_provision_mode,
                "cable_modem_status": cable_modem_status,
                "operator_id": operator_id,
            },
        )

        # return an enum-style metric for the provisioning status
        try:
            enum_provisioning_status = ProvisioningStatus(provisioning_status)
        except:
            raise ValueError(
                f"Unknown provisioning status '{provisioning_status}'. Please open an issue on Github."
            )
        yield StateSetMetricFamily(
            "connectbox_provisioning_status",
            "Provisioning status description",
            value={
                state.value: state == enum_provisioning_status
                for state in ProvisioningStatus
            },
        )

        # uptime is reported in a format like "36day(s)15h:24m:58s" which needs parsing
        uptime_pattern = r"(\d+)day\(s\)(\d+)h:(\d+)m:(\d+)s"
        m = re.fullmatch(uptime_pattern, uptime_as_str)
        if m is None:
            raise ValueError(
                f"Unexpected duration format '{uptime_as_str}', please open an issue on github."
            )
        uptime_timedelta = timedelta(
            days=int(m[1]), hours=int(m[2]), minutes=int(m[3]), seconds=int(m[4])
        )
        uptime_seconds = uptime_timedelta.total_seconds()

        yield GaugeMetricFamily(
            "connectbox_uptime",
            "Device uptime in seconds",
            unit="seconds",
            value=uptime_seconds,
        )


def get_metrics_extractor(ident: str):
    """
    Factory method for metrics extractors.
    :param ident: metric extractor identifier
    :return: extractor instance
    """
    extractors = {
        DEVICE_STATUS: DeviceStatusExtractor,
        DOWNSTREAM: DownstreamStatusExtractor,
        UPSTREAM: UpstreamStatusExtractor,
        LAN_USERS: LanUserExtractor,
        TEMPERATURE: TemperatureExtractor,
    }
    if not ident in extractors.keys():
        raise ValueError(
            f"Unknown extractor '{ident}', supported are: {', '.join(extractors.keys())}"
        )
    cls = extractors[ident]
    return cls()



from pathlib import Path
from typing import Union, Dict

from deepmerge import Merger
from ruamel.yaml import YAML

from connectbox_exporter.xml2metric import (
    DEVICE_STATUS,
    DOWNSTREAM,
    UPSTREAM,
    LAN_USERS,
    TEMPERATURE,
)

IP_ADDRESS = "ip_address"
PASSWORD = "password"
EXPORTER = "exporter"
PORT = "port"
TIMEOUT_SECONDS = "timeout_seconds"
EXTRACTORS = "metrics"

# pick default timeout one second less than the default prometheus timeout of 10s
DEFAULT_CONFIG = {
    EXPORTER: {
        PORT: 9705,
        TIMEOUT_SECONDS: 9,
        EXTRACTORS: {DEVICE_STATUS, DOWNSTREAM, UPSTREAM, LAN_USERS, TEMPERATURE},
    }
}


def load_config(config_file: Union[str, Path]) -> Dict:
    """
    Loads and validates YAML config for this exporter and fills in default values
    :param config_file:
    :return: config as dictionary
    """
    yaml = YAML()
    with open(config_file) as f:
        config = yaml.load(f)

    # merge with default config: use 'override' for lists to let users replace extractor setting entirely
    merger = Merger([(list, "override"), (dict, "merge")], ["override"], ["override"])
    config = merger.merge(DEFAULT_CONFIG, config)

    for param in [IP_ADDRESS, PASSWORD]:
        if not param in config:
            raise ValueError(
                f"'{param}' is a mandatory config parameter, but it is missing in the YAML configuration file. Please see README.md for an example."
            )
    if config[EXPORTER][TIMEOUT_SECONDS] <= 0:
        raise ValueError(f"'{TIMEOUT_SECONDS} must be positive.")
    if config[EXPORTER][PORT] < 0 or config[EXPORTER][PORT] > 65535:
        raise ValueError(f"Invalid exporter port.")

    if not config[EXPORTER][EXTRACTORS]:
        raise ValueError(
            "The config file needs to specify at least one family of metrics."
        )
    config[EXPORTER][EXTRACTORS] = sorted(set(config[EXPORTER][EXTRACTORS]))

    return config



from logging import Logger, NOTSET, addLevelName, INFO, DEBUG, StreamHandler, Formatter
import sys

VERBOSE = 5


class VerboseLogger(Logger):
    """
    Logger with custom log level VERBOSE which is lower than DEBUG.
    """

    def __init__(self, name, level=NOTSET):
        super().__init__(name, level)
        addLevelName(VERBOSE, "VERBOSE")

    def verbose(self, msg, *args, **kwargs):
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, msg, args, **kwargs)


def get_logger(verbosity: int) -> VerboseLogger:
    """
    Get logger object which logs to stdout with given verbosity.
    :param verbosity: logs INFO if 0, DEBUG if 1 and VERBOSE if >1
    :return:
    """
    if verbosity <= 0:
        log_level = INFO
    elif verbosity == 1:
        log_level = DEBUG
    else:
        log_level = VERBOSE

    logger = VerboseLogger("connectbox_exporter")
    logger.setLevel(log_level)
    handler = StreamHandler(sys.stdout)
    handler.setLevel(log_level)
    formatter = Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



import logging
import threading
import time
from typing import Dict
import json

import click
import compal
from lxml.etree import XMLSyntaxError
from prometheus_client import CollectorRegistry, MetricsHandler
from prometheus_client.exposition import _ThreadingSimpleServer
from prometheus_client.metrics_core import GaugeMetricFamily
from requests import Timeout

from connectbox_exporter.config import (
    load_config,
    IP_ADDRESS,
    PASSWORD,
    EXPORTER,
    PORT,
    TIMEOUT_SECONDS,
    EXTRACTORS,
)
from connectbox_exporter.logger import get_logger, VerboseLogger
from connectbox_exporter.xml2metric import get_metrics_extractor


class ConnectBoxCollector(object):
    def __init__(
        self,
        logger: VerboseLogger,
        ip_address: str,
        password: str,
        exporter_config: Dict,
    ):
        self.logger = logger
        self.ip_address = ip_address
        self.password = password
        self.timeout = exporter_config[TIMEOUT_SECONDS]

        extractors = exporter_config[EXTRACTORS]
        self.metric_extractors = [get_metrics_extractor(e) for e in extractors]

    def collect(self):
        # Collect scrape duration and scrape success for each extractor. Scrape success is initialized with False for
        # all extractors so that we can report a value for each extractor even in cases where we abort midway through
        # because we lost connection to the modem.
        scrape_duration = {}  # type: Dict[str, float]
        scrape_success = {e.name: False for e in self.metric_extractors}

        # attempt login
        login_logout_success = True
        try:
            self.logger.debug("Logging in at " + self.ip_address)
            connectbox = compal.Compal(
                self.ip_address, key=self.password, timeout=self.timeout
            )
            connectbox.login()
        except (ConnectionError, Timeout, ValueError) as e:
            self.logger.error(repr(e))
            connectbox = None
            login_logout_success = False

        # skip extracting further metrics if login failed
        if connectbox is not None:
            for extractor in self.metric_extractors:
                raw_xmls = {}
                try:
                    pre_scrape_time = time.time()

                    # obtain all raw XML responses for an extractor, then extract metrics
                    for fun in extractor.functions:
                        self.logger.debug(f"Querying fun={fun}...")
                        raw_xml = connectbox.xml_getter(fun, {}).content
                        self.logger.verbose(
                            f"Raw XML response for fun={fun}:\n{raw_xml.decode()}"
                        )
                        raw_xmls[fun] = raw_xml
                    yield from extractor.extract(raw_xmls)
                    post_scrape_time = time.time()

                    scrape_duration[extractor.name] = post_scrape_time - pre_scrape_time
                    scrape_success[extractor.name] = True
                except (XMLSyntaxError, AttributeError) as e:
                    # in case of a less serious error, log and continue scraping the next extractor
                    jsonized = json.dumps(raw_xmls)
                    message = f"Failed to extract '{extractor.name}'. Please open an issue on Github and include the following:\n{repr(e)}\n{jsonized}"
                    self.logger.error(message)
                except (ConnectionError, Timeout) as e:
                    # in case of serious connection issues, abort and do not try the next extractor
                    self.logger.error(repr(e))
                    break

            # attempt logout once done
            try:
                self.logger.debug("Logging out.")
                connectbox.logout()
            except Exception as e:
                self.logger.error(e)
                login_logout_success = False
        scrape_success["login_logout"] = int(login_logout_success)

        # create metrics from previously durations and successes collected
        EXTRACTOR = "extractor"
        scrape_duration_metric = GaugeMetricFamily(
            "connectbox_scrape_duration",
            documentation="Scrape duration by extractor",
            unit="seconds",
            labels=[EXTRACTOR],
        )
        for name, duration in scrape_duration.items():
            scrape_duration_metric.add_metric([name], duration)
        yield scrape_duration_metric

        scrape_success_metric = GaugeMetricFamily(
            "connectbox_up",
            documentation="Connect Box exporter scrape success by extractor",
            labels=[EXTRACTOR],
        )
        for name, success in scrape_success.items():
            scrape_success_metric.add_metric([name], int(success))
        yield scrape_success_metric


@click.command()
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "-v",
    "--verbose",
    help="Log more messages. Multiple -v increase verbosity.",
    count=True,
)
def main(config_file, verbose):
    """
    Launch the exporter using a YAML config file.
    """

    # hush the logger from the compal library and use our own custom logger
    compal.LOGGER.setLevel(logging.WARNING)
    logger = get_logger(verbose)

    # load user and merge with defaults
    config = load_config(config_file)
    exporter_config = config[EXPORTER]

    # fire up collector
    reg = CollectorRegistry()
    reg.register(
        ConnectBoxCollector(
            logger,
            ip_address=config[IP_ADDRESS],
            password=config[PASSWORD],
            exporter_config=config[EXPORTER],
        )
    )

    # start http server
    CustomMetricsHandler = MetricsHandler.factory(reg)
    httpd = _ThreadingSimpleServer(("", exporter_config[PORT]), CustomMetricsHandler)
    httpd_thread = threading.Thread(target=httpd.serve_forever)
    httpd_thread.start()

    logger.info(
        f"Exporter running at http://localhost:{exporter_config[PORT]}, querying {config[IP_ADDRESS]}"
    )

    # wait indefinitely
    try:
        while True:
            time.sleep(3)
    except KeyboardInterrupt:
        httpd.shutdown()
        httpd_thread.join()
